@@ -666,7 +666,7 @@ ssize_t _gnutls_io_write_flush(gnutls_session_t session)
  */
 int _gnutls_io_check_recv(gnutls_session_t session, unsigned int ms)
 {
-	gnutls_transport_ptr_t fd = session->internals.transport_send_ptr;
+	gnutls_transport_ptr_t fd = session->internals.transport_recv_ptr;
 	int ret = 0, err;
 
 	if (unlikely
@@ -924,7 +924,7 @@ static int merge_handshake_packet(gnutls_session_t session,
 	if (exists == 0)
 		pos = session->internals.handshake_recv_buffer_size;
 
-	if (pos > MAX_HANDSHAKE_MSGS)
+	if (pos >= MAX_HANDSHAKE_MSGS)
 		return
 		    gnutls_assert_val(GNUTLS_E_TOO_MANY_HANDSHAKE_PACKETS);
 
