@@ -168,11 +168,10 @@ static int resume_copy_required_values(gnutls_session_t session)
 	    NULL)
 		return gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR);
 
-	if (_gnutls_set_current_version(session,
+	_gnutls_set_current_version(session,
 				    session->internals.
 				    resumed_security_parameters.pversion->
-				    id) < 0)
-		return gnutls_assert_val(GNUTLS_E_UNSUPPORTED_VERSION_PACKET);
+				    id);
 
 	session->security_parameters.cert_type =
 	    session->internals.resumed_security_parameters.cert_type;
@@ -285,11 +284,11 @@ _gnutls_ssl3_finished(gnutls_session_t session, int type, uint8_t * ret,
 	else
 		len = session->internals.handshake_hash_buffer_prev_len;
 
-	rc = _gnutls_hash_init(&td_sha, mac_to_entry(GNUTLS_DIG_SHA1));
+	rc = _gnutls_hash_init(&td_sha, hash_to_entry(GNUTLS_DIG_SHA1));
 	if (rc < 0)
 		return gnutls_assert_val(rc);
 
-	rc = _gnutls_hash_init(&td_md5, mac_to_entry(GNUTLS_DIG_MD5));
+	rc = _gnutls_hash_init(&td_md5, hash_to_entry(GNUTLS_DIG_MD5));
 	if (rc < 0) {
 		_gnutls_hash_deinit(&td_sha, NULL);
 		return gnutls_assert_val(rc);
@@ -420,8 +419,7 @@ _gnutls_negotiate_version(gnutls_session_t session,
 		ret = adv_version;
 	}
 
-	if (_gnutls_set_current_version(session, ret) < 0)
-		return gnutls_assert_val(GNUTLS_E_UNSUPPORTED_VERSION_PACKET);
+	_gnutls_set_current_version(session, ret);
 
 	return ret;
 }
@@ -642,7 +640,7 @@ _gnutls_read_client_hello(gnutls_session_t session, uint8_t * data,
 		if (ret < 0)
 			return gnutls_assert_val(ret);
 
-		return 0;
+		return _gnutls_user_hello_func(session, adv_version);
 	}
 
 	/* select an appropriate cipher suite
@@ -679,7 +677,7 @@ static int _gnutls_send_finished(gnutls_session_t session, int again)
 
 	if (again == 0) {
 		bufel =
-		    _gnutls_handshake_alloc(session, MAX_VERIFY_DATA_SIZE,
+		    _gnutls_handshake_alloc(session, 
 					    MAX_VERIFY_DATA_SIZE);
 		if (bufel == NULL) {
 			gnutls_assert();
@@ -1152,7 +1150,7 @@ _gnutls_send_empty_handshake(gnutls_session_t session,
 	mbuffer_st *bufel;
 
 	if (again == 0) {
-		bufel = _gnutls_handshake_alloc(session, 0, 0);
+		bufel = _gnutls_handshake_alloc(session, 0);
 		if (bufel == NULL) {
 			gnutls_assert();
 			return GNUTLS_E_MEMORY_ERROR;
@@ -1728,8 +1726,7 @@ _gnutls_read_server_hello(gnutls_session_t session,
 		gnutls_assert();
 		return GNUTLS_E_UNSUPPORTED_VERSION_PACKET;
 	} else {
-		if (_gnutls_set_current_version(session, version) < 0)
-			return gnutls_assert_val(GNUTLS_E_UNSUPPORTED_VERSION_PACKET);
+		_gnutls_set_current_version(session, version);
 	}
 
 	pos += 2;
@@ -1747,7 +1744,7 @@ _gnutls_read_server_hello(gnutls_session_t session,
 	DECR_LEN(len, 1);
 	session_id_len = data[pos++];
 
-	if (len < session_id_len || session_id_len > TLS_MAX_SESSION_ID_SIZE) {
+	if (len < session_id_len) {
 		gnutls_assert();
 		return GNUTLS_E_UNSUPPORTED_VERSION_PACKET;
 	}
@@ -1960,8 +1957,7 @@ static int _gnutls_send_client_hello(gnutls_session_t session, int again)
 		 * (RSA uses it).
 		 */
 		set_adv_version(session, hver->major, hver->minor);
-		if (_gnutls_set_current_version(session, hver->id) < 0)
-			return gnutls_assert_val(GNUTLS_E_UNSUPPORTED_VERSION_PACKET);
+		_gnutls_set_current_version(session, hver->id);
 
 		if (session->internals.priorities.ssl3_record_version != 0) {
 			/* Advertize the SSL 3.0 record packet version in
@@ -2082,12 +2078,13 @@ static int _gnutls_send_client_hello(gnutls_session_t session, int again)
 		}
 
 		bufel =
-		    _gnutls_handshake_alloc(session, 0, extdata.length);
+		    _gnutls_handshake_alloc(session, extdata.length);
 		if (bufel == NULL) {
 			gnutls_assert();
 			ret = GNUTLS_E_MEMORY_ERROR;
 			goto cleanup; 
 		}
+		_mbuffer_set_udata_size(bufel, 0);
 
 		ret =
 		    _mbuffer_append_data(bufel, extdata.data,
@@ -2140,7 +2137,6 @@ static int _gnutls_send_server_hello(gnutls_session_t session, int again)
 
 		bufel =
 		    _gnutls_handshake_alloc(session,
-					    datalen + extdata.length,
 					    datalen + extdata.length);
 		if (bufel == NULL) {
 			gnutls_assert();
@@ -2355,8 +2351,8 @@ _gnutls_recv_hello_verify_request(gnutls_session_t session,
  * rehandshake is mandatory to terminate if it does not start after
  * some threshold.
  *
- * If the client does not wish to renegotiate parameters he will
- * should with an alert message, thus the return code will be
+ * If the client does not wish to renegotiate parameters he 
+ * should reply with an alert message, thus the return code will be
  * %GNUTLS_E_WARNING_ALERT_RECEIVED and the alert will be
  * %GNUTLS_A_NO_RENEGOTIATION.  A client may also choose to ignore
  * this message.
@@ -2424,7 +2420,7 @@ static int _gnutls_send_supplemental(gnutls_session_t session, int again)
 		}
 
 		bufel =
-		    _gnutls_handshake_alloc(session, buf.length,
+		    _gnutls_handshake_alloc(session, 
 					    buf.length);
 		if (bufel == NULL) {
 			gnutls_assert();
@@ -2481,26 +2477,18 @@ static int _gnutls_recv_supplemental(gnutls_session_t session)
  * has asked to resume a session, but the server couldn't, then a
  * full handshake will be performed.
  *
- * The non-fatal errors expected by this function are:
- * %GNUTLS_E_INTERRUPTED, %GNUTLS_E_AGAIN, 
- * %GNUTLS_E_WARNING_ALERT_RECEIVED, and %GNUTLS_GOT_APPLICATION_DATA,
- * the latter only in a case of rehandshake.
- *
- * The former two interrupt the handshake procedure due to the lower
- * layer being interrupted, and the latter because of an alert that
- * may be sent by a server (it is always a good idea to check any
- * received alerts). On these errors call this function again, until it
+ * The non-fatal errors such as %GNUTLS_E_AGAIN and
+ * %GNUTLS_E_INTERRUPTED interrupt the handshake procedure, which
+ * should be resumed later.  Call this function again, until it
  * returns 0; cf.  gnutls_record_get_direction() and
- * gnutls_error_is_fatal(). In DTLS sessions the non-fatal error
- * %GNUTLS_E_LARGE_PACKET is also possible, and indicates that
- * the MTU should be adjusted.
+ * gnutls_error_is_fatal().
  *
  * If this function is called by a server after a rehandshake request
  * then %GNUTLS_E_GOT_APPLICATION_DATA or
  * %GNUTLS_E_WARNING_ALERT_RECEIVED may be returned.  Note that these
  * are non fatal errors, only in the specific case of a rehandshake.
  * Their meaning is that the client rejected the rehandshake request or
- * in the case of %GNUTLS_E_GOT_APPLICATION_DATA it could also mean that
+ * in the case of %GNUTLS_E_GOT_APPLICATION_DATA it might also mean that
  * some data were pending.
  *
  * Returns: %GNUTLS_E_SUCCESS on success, otherwise a negative error code.
@@ -2512,16 +2500,17 @@ int gnutls_handshake(gnutls_session_t session)
 
 	/* sanity check. Verify that there are priorities setup.
 	 */
-	if (session->internals.priorities.protocol.algorithms == 0)
-		return gnutls_assert_val(GNUTLS_E_NO_PRIORITIES_WERE_SET);
 
 	if (STATE == STATE0) {
 		/* first call */
+		if (session->internals.priorities.protocol.algorithms == 0)
+			return gnutls_assert_val(GNUTLS_E_NO_PRIORITIES_WERE_SET);
+
 		gettime(&session->internals.dtls.handshake_start_time);
 		if (session->internals.handshake_timeout_ms &&
 		    session->internals.handshake_endtime == 0)
-		    session->internals.handshake_endtime = gnutls_time(0) +
-			    session->internals.handshake_timeout_ms / 1000;
+			    session->internals.handshake_endtime = session->internals.dtls.handshake_start_time.tv_sec +
+			    	session->internals.handshake_timeout_ms / 1000;
 	}
 
 	ret =
@@ -2580,10 +2569,7 @@ int gnutls_handshake(gnutls_session_t session)
  * to the provided value. Use an @ms value of zero to disable
  * timeout.
  *
- * Note that in order for the timeout to be enforced
- * gnutls_transport_set_pull_timeout_function() must be set
- * (it is set by default in most systems).
- *
+ * Since: 3.1.0
  **/
 void
 gnutls_handshake_set_timeout(gnutls_session_t session, unsigned int ms)
@@ -2599,18 +2585,10 @@ gnutls_handshake_set_timeout(gnutls_session_t session, unsigned int ms)
 		/* EAGAIN and INTERRUPTED are always non-fatal */ \
 		if (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED) \
 			return ret; \
-		if (ret == GNUTLS_E_GOT_APPLICATION_DATA && session->internals.initial_negotiation_completed != 0) \
-			return ret; \
-		if (ret == GNUTLS_E_LARGE_PACKET && session->internals.handshake_large_loops < 16) { \
-			session->internals.handshake_large_loops++; \
-			return ret; \
-		} \
                 /* a warning alert might interrupt handshake */ \
 		if (allow_alert != 0 && ret==GNUTLS_E_WARNING_ALERT_RECEIVED) return ret; \
 		gnutls_assert(); \
 		ERR( str, ret); \
-		if (gnutls_error_is_fatal(ret) == 0) ret = gnutls_assert_val(GNUTLS_E_INTERNAL_ERROR); \
-		session_invalidate(session); \
 		_gnutls_handshake_hash_buffers_clear(session); \
 		return ret; \
 	} } while (0)
@@ -2719,6 +2697,7 @@ static int _gnutls_handshake_client(gnutls_session_t session)
 		IMED_RET("recv server certificate", ret, 1);
 
 	case STATE6:
+#ifdef ENABLE_OCSP
 		/* RECV CERTIFICATE STATUS */
 		if (session->internals.resumed == RESUME_FALSE)	/* if we are not resuming */
 			ret =
@@ -2726,7 +2705,7 @@ static int _gnutls_handshake_client(gnutls_session_t session)
 			    (session);
 		STATE = STATE6;
 		IMED_RET("recv server certificate", ret, 1);
-
+#endif
 	case STATE7:
 		ret = run_verify_callback(session, GNUTLS_CLIENT);
 		STATE = STATE7;
@@ -2802,18 +2781,22 @@ static int _gnutls_handshake_client(gnutls_session_t session)
 		if (session->internals.resumed == RESUME_FALSE) {
 			ret = _gnutls_send_handshake_final(session, TRUE);
 			IMED_RET("send handshake final 2", ret, 1);
+#ifdef ENABLE_SESSION_TICKETS
 		} else {
 			ret = _gnutls_recv_new_session_ticket(session);
 			IMED_RET("recv handshake new session ticket", ret,
 				 1);
+#endif
 		}
 
 	case STATE16:
 		STATE = STATE16;
 		if (session->internals.resumed == RESUME_FALSE) {
+#ifdef ENABLE_SESSION_TICKETS
 			ret = _gnutls_recv_new_session_ticket(session);
 			IMED_RET("recv handshake new session ticket", ret,
 				 1);
+#endif
 		} else {
 			ret = _gnutls_recv_handshake_final(session, TRUE);
 			IMED_RET("recv handshake final", ret, 1);
@@ -2850,7 +2833,7 @@ static ssize_t send_change_cipher_spec(gnutls_session_t session, int again)
 	const version_entry_st *vers;
 
 	if (again == 0) {
-		bufel = _gnutls_handshake_alloc(session, 1, 1);
+		bufel = _gnutls_handshake_alloc(session, 1);
 		if (bufel == NULL)
 			return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
 
@@ -3091,6 +3074,7 @@ static int _gnutls_handshake_server(gnutls_session_t session)
 		IMED_RET("send server certificate", ret, 0);
 
 	case STATE4:
+#ifdef ENABLE_OCSP
 		if (session->internals.resumed == RESUME_FALSE)
 			ret =
 			    _gnutls_send_server_certificate_status(session,
@@ -3098,7 +3082,7 @@ static int _gnutls_handshake_server(gnutls_session_t session)
 								   (STATE4));
 		STATE = STATE4;
 		IMED_RET("send server certificate status", ret, 0);
-
+#endif
 	case STATE5:
 		/* send server key exchange (A) */
 		if (session->internals.resumed == RESUME_FALSE)
@@ -3175,12 +3159,13 @@ static int _gnutls_handshake_server(gnutls_session_t session)
 		}
 
 	case STATE13:
+#ifdef ENABLE_SESSION_TICKETS
 		ret =
 		    _gnutls_send_new_session_ticket(session,
 						    AGAIN(STATE13));
 		STATE = STATE13;
 		IMED_RET("send handshake new session ticket", ret, 0);
-
+#endif
 	case STATE14:
 		STATE = STATE14;
 		if (session->internals.resumed == RESUME_FALSE) {	/* if we are not resuming */
