@@ -65,7 +65,7 @@ struct node_st {
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
  *
- * Since: 3.0
+ * Since: 3.0.0
  **/
 int
 gnutls_x509_trust_list_init(gnutls_x509_trust_list_t * list,
@@ -73,6 +73,8 @@ gnutls_x509_trust_list_init(gnutls_x509_trust_list_t * list,
 {
 	gnutls_x509_trust_list_t tmp;
 
+	FAIL_IF_LIB_ERROR;
+
 	tmp =
 	    gnutls_calloc(1, sizeof(struct gnutls_x509_trust_list_st));
 
@@ -98,11 +100,14 @@ gnutls_x509_trust_list_init(gnutls_x509_trust_list_t * list,
 /**
  * gnutls_x509_trust_list_deinit:
  * @list: The structure to be deinitialized
- * @all: if non-(0) it will deinitialize all the certificates and CRLs contained in the structure.
+ * @all: if non-zero it will deinitialize all the certificates and CRLs contained in the structure.
  *
- * This function will deinitialize a trust list.
+ * This function will deinitialize a trust list. Note that the
+ * @all flag should be typically non-zero unless you have specified
+ * your certificates using gnutls_x509_trust_list_add_cas() and you
+ * want to prevent them from being deinitialized by this function.
  *
- * Since: 3.0
+ * Since: 3.0.0
  **/
 void
 gnutls_x509_trust_list_deinit(gnutls_x509_trust_list_t list,
@@ -119,64 +124,138 @@ gnutls_x509_trust_list_deinit(gnutls_x509_trust_list_t list,
 	gnutls_free(list->blacklisted);
 
 	for (i = 0; i < list->size; i++) {
-		if (all)
+		if (all) {
 			for (j = 0; j < list->node[i].trusted_ca_size; j++) {
 				gnutls_x509_crt_deinit(list->node[i].
 						       trusted_cas[j]);
 			}
+		}
 		gnutls_free(list->node[i].trusted_cas);
 
 
-		if (all)
+		if (all) {
 			for (j = 0; j < list->node[i].crl_size; j++) {
 				gnutls_x509_crl_deinit(list->node[i].
 						       crls[j]);
 			}
+		}
 		gnutls_free(list->node[i].crls);
 
-		if (all)
+		if (all) {
 			for (j = 0; j < list->node[i].named_cert_size; j++) {
 				gnutls_x509_crt_deinit(list->node[i].
 						       named_certs[j].
 						       cert);
 			}
+		}
 		gnutls_free(list->node[i].named_certs);
 	}
 
+	gnutls_free(list->x509_rdn_sequence.data);
 	gnutls_free(list->node);
 	gnutls_free(list->pkcs11_token);
 	gnutls_free(list);
 }
 
+static int
+add_new_ca_to_rdn_seq(gnutls_x509_trust_list_t list,
+		       gnutls_x509_crt_t ca)
+{
+	gnutls_datum_t tmp;
+	size_t newsize;
+	unsigned char *newdata, *p;
+
+	/* Add DN of the last added CAs to the RDN sequence
+	 * This will be sent to clients when a certificate
+	 * request message is sent.
+	 */
+
+	/* FIXME: in case of a client it is not needed
+	 * to do that. This would save time and memory.
+	 * However we don't have that information available
+	 * here.
+	 * Further, this function is now much more efficient,
+	 * so optimizing that is less important.
+	 */
+	tmp.data = ca->raw_dn.data;
+	tmp.size = ca->raw_dn.size;
+
+	newsize = list->x509_rdn_sequence.size + 2 + tmp.size;
+	if (newsize < list->x509_rdn_sequence.size) {
+		gnutls_assert();
+		return GNUTLS_E_SHORT_MEMORY_BUFFER;
+	}
+
+	newdata =
+	    gnutls_realloc_fast(list->x509_rdn_sequence.data,
+				newsize);
+	if (newdata == NULL) {
+		gnutls_assert();
+		return GNUTLS_E_MEMORY_ERROR;
+	}
+
+	p = newdata + list->x509_rdn_sequence.size;
+	_gnutls_write_uint16(tmp.size, p);
+	if (tmp.data != NULL)
+		memcpy(p + 2, tmp.data, tmp.size);
+
+	list->x509_rdn_sequence.size = newsize;
+	list->x509_rdn_sequence.data = newdata;
+
+	return 0;
+}
+
 /**
  * gnutls_x509_trust_list_add_cas:
  * @list: The structure of the list
  * @clist: A list of CAs
  * @clist_size: The length of the CA list
- * @flags: should be 0.
+ * @flags: should be 0 or an or'ed sequence of %GNUTLS_TL options.
  *
  * This function will add the given certificate authorities
  * to the trusted list. The list of CAs must not be deinitialized
  * during this structure's lifetime.
  *
+ * If the flag %GNUTLS_TL_NO_DUPLICATES is specified, then
+ * the provided @clist entries that are duplicates will not be
+ * added to the list and will be deinitialized.
+ *
  * Returns: The number of added elements is returned.
  *
- * Since: 3.0
+ * Since: 3.0.0
  **/
 int
 gnutls_x509_trust_list_add_cas(gnutls_x509_trust_list_t list,
 			       const gnutls_x509_crt_t * clist,
-			       int clist_size, unsigned int flags)
+			       unsigned clist_size, unsigned int flags)
 {
-	int i;
+	unsigned i, j;
 	uint32_t hash;
+	int ret;
+	unsigned exists;
 
 	for (i = 0; i < clist_size; i++) {
+		exists = 0;
 		hash =
 		    hash_pjw_bare(clist[i]->raw_dn.data,
 				  clist[i]->raw_dn.size);
 		hash %= list->size;
 
+		/* avoid duplicates */
+		if (flags & GNUTLS_TL_NO_DUPLICATES) {
+			for (j=0;j<list->node[hash].trusted_ca_size;j++) {
+				if (_gnutls_check_if_same_cert(list->node[hash].trusted_cas[j], clist[i]) != 0) {
+					exists = 1;
+					break;
+				}
+			}
+
+			if (exists != 0) {
+				gnutls_x509_crt_deinit(clist[i]);
+				continue;
+			}
+		}
+
 		list->node[hash].trusted_cas =
 		    gnutls_realloc_fast(list->node[hash].trusted_cas,
 					(list->node[hash].trusted_ca_size +
@@ -191,6 +270,14 @@ gnutls_x509_trust_list_add_cas(gnutls_x509_trust_list_t list,
 		list->node[hash].trusted_cas[list->node[hash].
 					     trusted_ca_size] = clist[i];
 		list->node[hash].trusted_ca_size++;
+
+		if (flags & GNUTLS_TL_USE_IN_TLS) {
+			ret = add_new_ca_to_rdn_seq(list, clist[i]);
+			if (ret < 0) {
+				gnutls_assert();
+				return i;
+			}
+		}
 	}
 
 	return i;
@@ -311,7 +398,7 @@ gnutls_x509_trust_list_remove_cas(gnutls_x509_trust_list_t list,
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
  *
- * Since: 3.0
+ * Since: 3.0.0
  **/
 int
 gnutls_x509_trust_list_add_named_crt(gnutls_x509_trust_list_t list,
@@ -429,8 +516,6 @@ gnutls_x509_trust_list_add_crls(gnutls_x509_trust_list_t list,
 /* Takes a certificate list and shortens it if there are
  * intermedia certificates already trusted by us.
  *
- * FIXME: This is very similar to _gnutls_x509_verify_certificate().
- *
  * Returns the new size of the list or a negative number on error.
  */
 static int shorten_clist(gnutls_x509_trust_list_t list,
@@ -543,22 +628,8 @@ static gnutls_x509_crt_t *sort_clist(gnutls_x509_crt_t
 	return sorted;
 }
 
-/**
- * gnutls_x509_trust_list_get_issuer:
- * @list: The structure of the list
- * @cert: is the certificate to find issuer for
- * @issuer: Will hold the issuer if any. Should be treated as constant.
- * @flags: Use (0).
- *
- * This function will attempt to find the issuer of the
- * given certificate.
- *
- * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
- *   negative error value.
- *
- * Since: 3.0
- **/
-int gnutls_x509_trust_list_get_issuer(gnutls_x509_trust_list_t list,
+static
+int trust_list_get_issuer(gnutls_x509_trust_list_t list,
 				      gnutls_x509_crt_t cert,
 				      gnutls_x509_crt_t * issuer,
 				      unsigned int flags)
@@ -586,6 +657,52 @@ int gnutls_x509_trust_list_get_issuer(gnutls_x509_trust_list_t list,
 	return GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
 }
 
+/**
+ * gnutls_x509_trust_list_get_issuer:
+ * @list: The structure of the list
+ * @cert: is the certificate to find issuer for
+ * @issuer: Will hold the issuer if any. Should be treated as constant.
+ * @flags: Use zero.
+ *
+ * This function will attempt to find the issuer of the
+ * given certificate.
+ *
+ * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
+ *   negative error value.
+ *
+ * Since: 3.0
+ **/
+int gnutls_x509_trust_list_get_issuer(gnutls_x509_trust_list_t list,
+				      gnutls_x509_crt_t cert,
+				      gnutls_x509_crt_t * issuer,
+				      unsigned int flags)
+{
+	int ret;
+
+#ifdef ENABLE_PKCS11
+	if (list->pkcs11_token) {
+		gnutls_datum_t der = {NULL, 0};
+		/* use the token for verification */
+		ret = gnutls_pkcs11_get_raw_issuer(list->pkcs11_token, cert, &der,
+			GNUTLS_X509_FMT_DER, 0);
+		if (ret < 0)
+			return gnutls_assert_val(ret);
+
+		/* we add this CA to the trusted list in order to make it
+		 * persistent. It will be deallocated when the trust list is.
+		 */
+		ret = gnutls_x509_trust_list_add_trust_mem(list, &der, NULL,
+			GNUTLS_X509_FMT_DER, GNUTLS_TL_NO_DUPLICATES, 0);
+		if (ret < 0)
+			return gnutls_assert_val(ret);
+
+		gnutls_free(der.data);
+	}
+#endif
+
+	return trust_list_get_issuer(list, cert, issuer, flags);
+}
+
 static
 int check_if_in_blacklist(gnutls_x509_crt_t * cert_list, unsigned int cert_list_size,
 	gnutls_x509_crt_t * blacklist, unsigned int blacklist_size)
@@ -612,13 +729,18 @@ unsigned i, j;
  * @cert_list: is the certificate list to be verified
  * @cert_list_size: is the certificate list size
  * @flags: Flags that may be used to change the verification algorithm. Use OR of the gnutls_certificate_verify_flags enumerations.
- * @verify: will hold the certificate verification output.
+ * @voutput: will hold the certificate verification output.
  * @func: If non-null will be called on each chain element verification with the output.
  *
  * This function will try to verify the given certificate and return
  * its status. The @verify parameter will hold an OR'ed sequence of
  * %gnutls_certificate_status_t flags.
  *
+ * Additionally a certificate verification profile can be specified
+ * from the ones in %gnutls_certificate_verification_profiles_t by
+ * ORing the result of GNUTLS_PROFILE_TO_VFLAGS() to the verification
+ * flags.
+ *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
  *
@@ -629,7 +751,7 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 				  gnutls_x509_crt_t * cert_list,
 				  unsigned int cert_list_size,
 				  unsigned int flags,
-				  unsigned int *verify,
+				  unsigned int *voutput,
 				  gnutls_verify_output_function func)
 {
 	int ret;
@@ -657,23 +779,23 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 	ret = check_if_in_blacklist(cert_list, cert_list_size,
 		list->blacklisted, list->blacklisted_size);
 	if (ret != 0) {
-		*verify |= GNUTLS_CERT_REVOKED;
-		*verify |= GNUTLS_CERT_INVALID;
+		*voutput |= GNUTLS_CERT_REVOKED;
+		*voutput |= GNUTLS_CERT_INVALID;
 		return 0;
 	}
 
 #ifdef ENABLE_PKCS11
 	if (list->pkcs11_token) {
-		/* use the token for verification */	
+		/* use the token for verification */
 
-		*verify = _gnutls_pkcs11_verify_certificate(list->pkcs11_token,
+		*voutput = _gnutls_pkcs11_verify_crt_status(list->pkcs11_token,
 								cert_list, cert_list_size,
 								flags, func);
 	} else
 #endif
 	{
-		*verify =
-		    _gnutls_x509_verify_certificate(cert_list, cert_list_size,
+		*voutput =
+		    _gnutls_verify_crt_status(cert_list, cert_list_size,
 						    list->node[hash].trusted_cas,
 						    list->
 						    node[hash].trusted_ca_size,
@@ -682,7 +804,7 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 #define LAST_DN cert_list[cert_list_size-1]->raw_dn
 #define LAST_IDN cert_list[cert_list_size-1]->raw_issuer_dn
 
-		if ((*verify) & GNUTLS_CERT_SIGNER_NOT_FOUND &&
+		if ((*voutput) & GNUTLS_CERT_SIGNER_NOT_FOUND &&
 			(LAST_DN.size != LAST_IDN.size ||
 			 memcmp(LAST_DN.data, LAST_IDN.data, LAST_IDN.size) != 0)) {
 
@@ -694,8 +816,8 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 				  data, cert_list[cert_list_size - 1]->raw_dn.size);
 			hash %= list->size;
 
-			*verify =
-			    _gnutls_x509_verify_certificate(cert_list, cert_list_size,
+			*voutput =
+			    _gnutls_verify_crt_status(cert_list, cert_list_size,
 						    list->node[hash].trusted_cas,
 						    list->
 						    node[hash].trusted_ca_size,
@@ -703,7 +825,7 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 		}
 	}
 
-	if (*verify != 0 || (flags & GNUTLS_VERIFY_DISABLE_CRL_CHECKS))
+	if (*voutput != 0 || (flags & GNUTLS_VERIFY_DISABLE_CRL_CHECKS))
 		return 0;
 
 	/* Check revocation of individual certificates.
@@ -716,8 +838,8 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 					      list->node[hash].crl_size,
 					      func);
 	if (ret == 1) {		/* revoked */
-		*verify |= GNUTLS_CERT_REVOKED;
-		*verify |= GNUTLS_CERT_INVALID;
+		*voutput |= GNUTLS_CERT_REVOKED;
+		*voutput |= GNUTLS_CERT_INVALID;
 		return 0;
 	}
 
@@ -732,9 +854,11 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
 							crls,
 							list->node[hash].
 							crl_size, func);
-		if (ret == 1) {	/* revoked */
-			*verify |= GNUTLS_CERT_REVOKED;
-			*verify |= GNUTLS_CERT_INVALID;
+		if (ret < 0) {
+			gnutls_assert();
+		} else if (ret == 1) {	/* revoked */
+			*voutput |= GNUTLS_CERT_REVOKED;
+			*voutput |= GNUTLS_CERT_INVALID;
 			return 0;
 		}
 	}
@@ -749,18 +873,23 @@ gnutls_x509_trust_list_verify_crt(gnutls_x509_trust_list_t list,
  * @name: is the certificate's name
  * @name_size: is the certificate's name size
  * @flags: Flags that may be used to change the verification algorithm. Use OR of the gnutls_certificate_verify_flags enumerations.
- * @verify: will hold the certificate verification output.
+ * @voutput: will hold the certificate verification output.
  * @func: If non-null will be called on each chain element verification with the output.
  *
  * This function will try to find a certificate that is associated with the provided
  * name --see gnutls_x509_trust_list_add_named_crt(). If a match is found the certificate is considered valid. 
  * In addition to that this function will also check CRLs. 
- * The @verify parameter will hold an OR'ed sequence of %gnutls_certificate_status_t flags.
+ * The @voutput parameter will hold an OR'ed sequence of %gnutls_certificate_status_t flags.
+ *
+ * Additionally a certificate verification profile can be specified
+ * from the ones in %gnutls_certificate_verification_profiles_t by
+ * ORing the result of GNUTLS_PROFILE_TO_VFLAGS() to the verification
+ * flags.
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
  *
- * Since: 3.0
+ * Since: 3.0.0
  **/
 int
 gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
@@ -768,7 +897,7 @@ gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
 					const void *name,
 					size_t name_size,
 					unsigned int flags,
-					unsigned int *verify,
+					unsigned int *voutput,
 					gnutls_verify_output_function func)
 {
 	int ret;
@@ -783,12 +912,12 @@ gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
 	ret = check_if_in_blacklist(&cert, 1,
 		list->blacklisted, list->blacklisted_size);
 	if (ret != 0) {
-		*verify |= GNUTLS_CERT_REVOKED;
-		*verify |= GNUTLS_CERT_INVALID;
+		*voutput |= GNUTLS_CERT_REVOKED;
+		*voutput |= GNUTLS_CERT_INVALID;
 		return 0;
 	}
 
-	*verify = GNUTLS_CERT_INVALID | GNUTLS_CERT_SIGNER_NOT_FOUND;
+	*voutput = GNUTLS_CERT_INVALID | GNUTLS_CERT_SIGNER_NOT_FOUND;
 
 	for (i = 0; i < list->node[hash].named_cert_size; i++) {
 		if (_gnutls_check_if_same_cert(cert, list->node[hash].named_certs[i].cert) != 0) {	/* check if name matches */
@@ -796,13 +925,13 @@ gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
 			    name_size
 			    && memcmp(list->node[hash].named_certs[i].name,
 				      name, name_size) == 0) {
-				*verify = 0;
+				*voutput = 0;
 				break;
 			}
 		}
 	}
 
-	if (*verify != 0 || (flags & GNUTLS_VERIFY_DISABLE_CRL_CHECKS))
+	if (*voutput != 0 || (flags & GNUTLS_VERIFY_DISABLE_CRL_CHECKS))
 		return 0;
 
 	/* Check revocation of individual certificates.
@@ -813,8 +942,8 @@ gnutls_x509_trust_list_verify_named_crt(gnutls_x509_trust_list_t list,
 						list->node[hash].crl_size,
 						func);
 	if (ret == 1) {		/* revoked */
-		*verify |= GNUTLS_CERT_REVOKED;
-		*verify |= GNUTLS_CERT_INVALID;
+		*voutput |= GNUTLS_CERT_REVOKED;
+		*voutput |= GNUTLS_CERT_INVALID;
 		return 0;
 	}
 
